#code "bones multiTexturing pointCookies pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "bones pointCookies pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "multiTexturing pointCookies pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "pointCookies pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "bones multiTexturing pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = vec4(1.0); // Default
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "bones pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = vec4(1.0); // Default
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "multiTexturing pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = vec4(1.0); // Default
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "pointLight pointShadows vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_shadowTexture;
uniform float R_shadowMinVariance;
uniform float R_lightBleedingReduction;
uniform float R_cubeEdgeSize;
uniform float R_shadowMapNear;
uniform float R_shadowMapFar;
float SampleShadowMap(sampler2D shadowMap, vec2 coords, float compare)
{
	return step(compare, texture2D(shadowMap, coords).r);
}
float linstep(float low, float high, float v)
{
	return clamp((v-low)/(high-low), 0.0, 1.0);
}
float SampleVarianceShadowMap(sampler2D shadowMap, vec2 coords, float compare, float varianceMin, float lightBleedingReduction)
{
	vec2 moments = texture2D(shadowMap, coords).xy;
	float p = step(compare, moments.x);
	float variance = max(moments.y - moments.x * moments.x, varianceMin);
	float d = compare - moments.x;
	float pMax = linstep(lightBleedingReduction, 1.0, variance / (variance + d*d));
	return min(max(p, pMax), 1.0);
}
float SampleShadowMapLinear(sampler2D shadowMap, vec2 coords, float compare, vec2 texelSize)
{
	vec2 pixelPos = coords/texelSize + vec2(0.5);
	vec2 fracPart = fract(pixelPos);
	vec2 startTexel = (pixelPos - fracPart) * texelSize;
	float blTexel = SampleShadowMap(shadowMap, startTexel, compare);
	float brTexel = SampleShadowMap(shadowMap, startTexel + vec2(texelSize.x, 0.0), compare);
	float tlTexel = SampleShadowMap(shadowMap, startTexel + vec2(0.0, texelSize.y), compare);
	float trTexel = SampleShadowMap(shadowMap, startTexel + texelSize, compare);
	float mixA = mix(blTexel, tlTexel, fracPart.y);
	float mixB = mix(brTexel, trTexel, fracPart.y);
	return mix(mixA, mixB, fracPart.x);
}
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
float VectorToDepthValue(vec3 Vec)
{
	vec3 AbsVec = abs(Vec);
	float LocalZcomp = max(AbsVec.x, max(AbsVec.y, AbsVec.z)) - 0.01;
	float f = R_shadowMapFar;
	float n = R_shadowMapNear;
	float NormZComp = (f+n) / (f-n) - (2*f*n)/(f-n)/LocalZcomp;
	return (NormZComp + 1.0) * 0.5;
}
float SamplePointShadowMap(sampler2D shadowMap, vec3 directionToLight)
{
	return SampleVarianceShadowMap(shadowMap, GetCubeTC(directionToLight,R_cubeEdgeSize,R_cubeEdgeSize), VectorToDepthValue(directionToLight), R_shadowMinVariance, R_lightBleedingReduction);
}
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = SamplePointShadowMap(R_shadowTexture, outPosition - lightPosition);
	vec4 cookieColor = vec4(1.0); // Default
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "bones multiTexturing pointCookies pointLight vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = 1.0;
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "bones pointCookies pointLight vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
uniform mat4 boneMatrix[64];
attribute vec4 boneId;
attribute vec4 boneWeight;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	mat4 boneTransform = boneMatrix[int(boneId.x)] * boneWeight.x;
	boneTransform += boneMatrix[int(boneId.y)] * boneWeight.y;
	boneTransform += boneMatrix[int(boneId.z)] * boneWeight.z;
	boneTransform += boneMatrix[int(boneId.w)] * boneWeight.w;
	vec4 vertexPosition = boneTransform * vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = boneTransform * vec4(normal, 0.0);
	vec4 vertexTangent = boneTransform * vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = 1.0;
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "multiTexturing pointCookies pointLight vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}
uniform vec4 uvAnimation0;
uniform vec4 uvAnimation1;
uniform vec4 uvAnimation2;
uniform vec4 uvAnimation3;
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
	outMultiUV0 = vec4(UV * uvAnimation0.zw + uvAnimation0.xy, UV * uvAnimation1.zw + uvAnimation1.xy);
	outMultiUV1 = vec4(UV * uvAnimation2.zw + uvAnimation2.xy, UV * uvAnimation3.zw + uvAnimation3.xy);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
varying vec4 outMultiUV0;
varying vec4 outMultiUV1;
uniform sampler2D diffuseTexture0;
uniform sampler2D diffuseTexture1;
uniform sampler2D diffuseTexture2;
uniform sampler2D diffuseTexture3;
uniform sampler2D alphaTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = 1.0;
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 alphaColor = texture2D(alphaTexture, uvCoords);
	vec4 textureColor = texture2D(diffuseTexture0, outMultiUV0.xy) * alphaColor.r;
	textureColor += texture2D(diffuseTexture1, outMultiUV0.zw) * alphaColor.g;
	textureColor += texture2D(diffuseTexture2, outMultiUV1.xy) * alphaColor.b;
	textureColor += texture2D(diffuseTexture3, outMultiUV1.zw) * alphaColor.a;
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
#code "pointCookies pointLight vertexLighting"
#shader vertex
varying vec2 outUV;
varying vec3 outPosition;
uniform mat4 T_modelViewProjectionMatrix;
uniform mat4 T_modelMatrix;
uniform vec4 uvAnimation;
attribute vec2 UV;
attribute vec3 vertex;
varying vec3 outNormal;
uniform mat3 T_normalMatrix;
attribute vec3 normal;
attribute vec4 tangent;
uniform vec4 diffuse;
uniform vec4 specular;
uniform float shininess;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
struct DirectionalLight
{
	BaseLight base;
	vec3 direction;
};
struct SpotLight
{
	PointLight pointLight;
	vec3 direction;
	float cutoff;
};
uniform vec3 T_eyePos;
vec4 CalcLight(BaseLight base, vec3 surf2light, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 norm = normalize(normal);
	float dcont = max(0.0, dot(norm, surf2light));
	vec4 diffuse = dcont * texColor * diffuse * vec4(base.diffuse, 0.0);
	vec3 surf2view = normalize(T_eyePos-worldPos);
	vec3 refl = normalize(reflect(-surf2light, norm));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess * specFactor.a);
	vec4 specular = scont * specular * vec4(base.specular * specFactor.rgb, 0.0);
	return diffuse+specular;
}
vec4 CalcPointLight(PointLight pointLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = pointLight.position-worldPos;
	float d = length(lightDirection);
	if(d>pointLight.range)
	return vec4(0.0,0.0,0.0,0.0);
	float att = (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001);
	return CalcLight(pointLight.base, normalize(lightDirection),texColor,specFactor,worldPos,normal)/att;
}
vec4 CalcDirectionalLight(DirectionalLight directionalLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	return CalcLight(directionalLight.base, -directionalLight.direction, texColor, specFactor, worldPos, normal);
}
vec4 CalcSpotLight(SpotLight spotLight, vec4 texColor, vec4 specFactor, vec3 worldPos, vec3 normal)
{
	vec3 lightDirection = normalize(worldPos - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	vec4 color = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		color = CalcPointLight(spotLight.pointLight, texColor, specFactor, worldPos, normal) * (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
	}
	return color;
}
varying vec4 outDiffuse;
varying vec4 outSpecular;
void CalcVertexLight(BaseLight base, vec3 vertexPosition, vec3 vertexNormal, vec3 lightDirection, float att)
{
	float dcont = max(0.0, dot(vertexNormal, lightDirection));
	outDiffuse = dcont * diffuse * vec4(base.diffuse,1.0) / att;
	vec3 surf2view = normalize(T_eyePos-vertexPosition);
	vec3 refl = normalize(reflect(-lightDirection, vertexNormal));
	float scont = pow(max(0.0, dot(surf2view, refl)), shininess);
	outSpecular = scont * specular * vec4(base.specular,1.0) / att;
}
void CalcVertexPointLight(PointLight pointLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = pointLight.position-vertexPosition;
	float d = length(lightDirection);
	if(d<=pointLight.range)
	{
		CalcVertexLight(pointLight.base, vertexPosition, vertexNormal, normalize(lightDirection), (pointLight.attenuation.x+pointLight.attenuation.y*d+pointLight.attenuation.z*d*d+0.000001));
	}
	else
	{
		outDiffuse = vec4(0.0,0.0,0.0,1.0);
		outSpecular = vec4(0.0,0.0,0.0,1.0);
	}
}
void CalcVertexDirectionalLight(DirectionalLight directionalLight, vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexLight(directionalLight.base, vertexPosition, vertexNormal, -directionalLight.direction, 1.0);
}
void CalcVertexSpotLight(SpotLight spotLight, vec3 vertexPosition, vec3 vertexNormal)
{
	vec3 lightDirection = normalize(vertexPosition - spotLight.pointLight.position);
	float spotFactor = dot(lightDirection, spotLight.direction);
	outDiffuse = vec4(0.0,0.0,0.0,0.0);
	outSpecular = vec4(0.0,0.0,0.0,0.0);
	if(spotFactor > spotLight.cutoff)
	{
		CalcVertexPointLight(spotLight.pointLight, vertexPosition, vertexNormal);
		float spot = (1.0 - (1.0 - spotFactor)/(1.0 - spotLight.cutoff));
		outDiffuse *= spot;
		outSpecular *= spot;
	}
}
uniform PointLight R_light;
void CalcVertexLightingEffect(vec3 vertexPosition, vec3 vertexNormal)
{
	CalcVertexPointLight(R_light, vertexPosition, vertexNormal);
}

void main()
{
	vec4 vertexPosition = vec4(vertex, 1.0);
	gl_Position = T_modelViewProjectionMatrix * vertexPosition;
	outUV = UV * uvAnimation.zw + uvAnimation.xy;
	outPosition=vec3(T_modelMatrix * vertexPosition);
	vec4 vertexNormal = vec4(normal, 0.0);
	vec4 vertexTangent = vec4(tangent.xyz, 0.0);
	outNormal=normalize(T_normalMatrix * vertexNormal.xyz);
	CalcVertexLightingEffect(outPosition, outNormal);
}

#shader fragment
varying vec2 outUV;
varying vec3 outPosition;
uniform vec4 R_clipPlane;
varying vec3 outNormal;
varying vec4 outDiffuse;
varying vec4 outSpecular;
struct BaseLight
{
	vec3 diffuse;
	vec3 specular;
};
struct PointLight
{
	BaseLight base;
	vec3 position;
	vec3 attenuation;
	float range;
};
uniform PointLight R_light;
uniform sampler2D R_cookieTexture;
uniform vec4 R_uvAnimation;
uniform float R_cubeEdgeSize;
vec2 GetCubeTC(vec3 Dir,float xOffset,float yOffset)
{
	float Sc;
	float Tc;
	float Ma;
	float FaceIndex;
	float rx = Dir.x;
	float ry = Dir.y;
	float rz = Dir.z;
	vec3 adir = abs(Dir);
	Ma = max(max(adir.x, adir.y), adir.z);
	if (adir.x > adir.y && adir.x > adir.z)
	{
		Sc = (rx > 0.0) ? rz : -rz;
		Tc = ry;
		FaceIndex = (rx > 0.0) ? 0.0 : 1.0;
	}
	else if (adir.y > adir.x && adir.y > adir.z)
	{
		Sc = rx;
		Tc = (ry > 0.0) ? rz : -rz;
		FaceIndex = (ry > 0.0) ? 2.0 : 3.0;
	}
	else
	{
		Sc = (rz > 0.0) ? -rx : rx;
		Tc = ry;
		FaceIndex = (rz > 0.0) ? 4.0 : 5.0;
	}
	float s = 0.5 * (Sc / Ma + 1.0);
	float t = 0.5 * (Tc / Ma + 1.0);
	s = xOffset + (1.0-2*xOffset) * s;
	t = yOffset + (1.0-2*yOffset) * t;
	s = s / 3.0;
	t = t / 2.0;
	float Flr = floor(FaceIndex / 3.0);
	float Rmd = FaceIndex - (3.0 * Flr);
	s += Rmd / 3.0;
	t += Flr / 2.0;
	return vec2(s, t);
}
uniform sampler2D diffuseTexture;
uniform sampler2D specularTexture;

void main()
{
	if(dot(outPosition, R_clipPlane.xyz) < R_clipPlane.w)
	discard;
	vec3 lightPosition = R_light.position;
	float shadowFactor = 1.0;
	vec4 cookieColor = texture2D(R_cookieTexture, GetCubeTC(outPosition - lightPosition,R_cubeEdgeSize,R_cubeEdgeSize) * R_uvAnimation.zw + R_uvAnimation.xy);
	vec2 uvCoords = outUV; //input
	vec4 textureColor = vec4(1.0); //output
	textureColor = texture2D(diffuseTexture, uvCoords);
	vec4 specularColor = vec4(1.0); //output
	specularColor = texture2D(specularTexture, uvCoords);
	gl_FragColor = shadowFactor * cookieColor * (textureColor * outDiffuse + outSpecular * vec4(specularColor.rgb, 1.0));
}

#endcode
